@@ -16,6 +16,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -435,10 +457,10 @@ parse_variable (gst_parser *p)
 }
 
 
-/* attributes: attributes '<' attribute '>'
+/* attributes: attributes '<' attribute_keywords '>'
 	     | empty
 
-   attribute: attribute KEYWORD binary_expr 
+   attribute_keywords: attribute KEYWORD binary_expr 
 	    | KEYWORD binary_expr */
 
 static tree_node
@@ -625,11 +647,12 @@ parse_primary (gst_parser *p)
 		   | IDENTIFIER  */
 
 static tree_node
-parse_variable_primary (gst_parser *p)
+parse_variable_primary_1 (gst_parser *p, YYLTYPE *first_loc,
+			  const char *first_val)
 {
   tree_node node;
   assert (p->token == IDENTIFIER);
-  node = _gst_make_variable (&p->loc, p->val.sval);
+  node = _gst_make_variable (first_loc, first_val);
   for (;;)
     {
       lex (p);
@@ -643,6 +666,12 @@ parse_variable_primary (gst_parser *p)
   return node;
 }
 
+static tree_node
+parse_variable_primary (gst_parser *p)
+{
+  return parse_variable_primary_1 (p, &p->loc, p->val.sval);
+}
+
 
 /* array_literal_elt: array_literal
 		    | byte_array_literal
