@@ -19,6 +19,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -147,7 +169,7 @@ typedef struct VMProxy
   mst_Boolean (*respondsTo) __PROTO ((OOP oop, OOP selector));
   size_t (*OOPSize) __PROTO ((OOP oop));
   OOP (*OOPAt) __PROTO ((OOP oop, size_t index));
-  OOP (*OOPAtPut) __PROTO ((OOP oop, size_t index, OOP new));
+  OOP (*OOPAtPut) __PROTO ((OOP oop, size_t index, OOP newOOP));
 
   /* Some system classes.  */
   OOP objectClass, arrayClass, stringClass, characterClass, smallIntegerClass,
@@ -170,8 +192,8 @@ typedef struct VMProxy
 #define INDEXED_WORD(obj, n)   ( ((long *) ((obj) + 1))		    [(n)-1] )
 #define INDEXED_BYTE(obj, n)   ( ((char *) ((obj) + 1))		    [(n)-1] )
 #define INDEXED_OOP(obj, n)    ( ((OOP  *) ((obj) + 1))		    [(n)-1] )
-#define ARRAY_OOP_AT(obj, n)   ( ((OOP  *) ((mst_Object) obj)->data) [(n)-1] )
-#define STRING_OOP_AT(obj, n)  ( ((char *) ((mst_Object) obj)->data) [(n)-1] )
+#define ARRAY_OOP_AT(obj, n)   ( ((OOP  *) ((gst_object) obj)->data) [(n)-1] )
+#define STRING_OOP_AT(obj, n)  ( ((char *) ((gst_object) obj)->data) [(n)-1] )
 
 /* Compatibility section */
 #define indexedWord(obj, n)   INDEXED_WORD(obj, n)
