@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2004,2005,2006,2007
+ * Copyright 1988,89,90,91,92,94,95,99,2000,2001,2002,2003,2004,2005,2006
  * Free Software Foundation, Inc.
  * Written by Steve Byrne.
  *
@@ -98,21 +98,23 @@ static mst_Boolean is_base_digit (int c,
 				  int base);
 
 /* Parse the fractional part of a Float constant.  Store it in
-   NUMPTR.  Read numbers in base-BASE, the first one being C.  Answer the
-   scale (number of digits in numPtr).  If LARGEINTEGER is not NULL,
-   the digits are stored in an obstack, and LARGEINTEGER is set to true
-   if numPtr does not have sufficient precision.  */
+   NUMPTR.  Read numbers in base-BASE, the first one being C. 
+   Answer the scale (number of digits in numPtr).  If LARGEINTEGER
+   is not NULL, the digits are stored in the buffer maintained
+   by str.c, and LARGEINTEGER is set to true if numPtr does
+   not have sufficient precision.  */
 static int scan_fraction (int c,
 			   int base,
 			   long double *numPtr,
 			   mst_Boolean *largeInteger);
 
 /* Parse a numeric constant and return it.  Read numbers in
-   base-BASE, the first one being C.  If a - was parsed, NEGATIVE
-   must be true so that the sign of the result is changed accordingly.
-   If LARGEINTEGER is not NULL, the digits are stored in an obstack,
-   and LARGEINTEGER is set to true if the return value does not have
-   sufficient precision.  */
+   base-BASE, the first one being C.  If a - was parsed,
+   NEGATIVE must be true so that the sign of the result is
+   changed accordingly.  If LARGEINTEGER is not NULL, the
+   digits are stored in the buffer maintained by str.c,
+   and LARGEINTEGER is set to true if the return value does
+   not have sufficient precision.  */
 static long double scan_digits (int c,
 			    mst_Boolean negative,
 			    int base,
@@ -140,14 +142,7 @@ static int char_literal (int c,
 
 /* Parse a binary operator.  C is the first symbol in the selector */
 static int scan_bin_op (int c,
-			YYSTYPE * lvalp);
-
-/* Actual work for scan_bin_op is done here.  MAYBE_NUMBER is false if
-   we cannot parse a negative number in this context.  */ 
-static int scan_bin_op_1 (int c,
-			  YYSTYPE * lvalp,
-			  mst_Boolean maybe_number);
-
+			 YYSTYPE * lvalp);
 
 /* Parse a string literal.  C is '\'' */
 static int string_literal (int c,
@@ -174,10 +169,9 @@ static int scan_symbol (int c,
 static int digit_to_int (int c,
 			 int base);
 
-/* Raise BASE to the N-th power and multiply MANT by the result.  */
-static inline long double mul_powl (long double mant,
-				    long double base,
-			            int n);
+/* Raise BASE to the N-th power.  */
+static inline long double ipowl (long double base,
+			         int n);
 
 #ifdef LEXDEBUG
 static void print_token (int token,
@@ -494,7 +488,7 @@ scan_symbol (int c,
   /* We can read a binary operator and return a SYMBOL_LITERAL,... */
   if (CHAR_TAB (ic)->char_class & BIN_OP_CHAR)
     {
-      scan_bin_op_1 (ic, lvalp, false);
+      scan_bin_op (ic, lvalp);
       return SYMBOL_LITERAL;
     }
 
@@ -512,23 +506,22 @@ scan_symbol (int c,
       return '#';
     }
 
-  obstack_1grow (_gst_compilation_obstack, ic);
+  _gst_reset_buffer ();
+  _gst_add_str_buf_char (ic);
 
   while (((ic = _gst_next_char ()) != EOF)
          && (CHAR_TAB (ic)->char_class & SYMBOL_CHAR))
-    obstack_1grow (_gst_compilation_obstack, ic);
+    _gst_add_str_buf_char (ic);
 
   _gst_unread_char (ic);
-  obstack_1grow (_gst_compilation_obstack, '\0');
-  lvalp->sval = obstack_finish (_gst_compilation_obstack);
+  lvalp->sval = _gst_obstack_cur_str_buf (_gst_compilation_obstack);
   return SYMBOL_LITERAL;
 }
 
 
 int
-scan_bin_op_1 (int c,
-	       YYSTYPE *lvalp,
-	       mst_Boolean maybe_number)
+scan_bin_op (int c,
+	      YYSTYPE *lvalp)
 {
   char buf[3];
   int ic;
@@ -559,7 +552,7 @@ scan_bin_op_1 (int c,
 
       /* We come here also for a negative number, which we handle
          specially.  */
-      if (maybe_number && c == '-' && is_digit (ic))
+      if (c == '-' && is_digit (ic))
 	return (scan_number ('-', lvalp));
 
       buf[1] = 0;
@@ -575,19 +568,14 @@ scan_bin_op_1 (int c,
     return (BINOP);
 }
 
-int
-scan_bin_op (int c,
-	     YYSTYPE *lvalp)
-{
-  return scan_bin_op_1 (c, lvalp, true);
-}
-
 int
 string_literal (int c,
 		YYSTYPE * lvalp)
 {
   int ic;
 
+  _gst_reset_buffer ();
+
   for (;;)
     {
       ic = _gst_next_char ();
@@ -607,10 +595,9 @@ string_literal (int c,
 	      break;
 	    }
 	}
-      obstack_1grow (_gst_compilation_obstack, ic);
+      _gst_add_str_buf_char (ic);
     }
-  obstack_1grow (_gst_compilation_obstack, '\0');
-  lvalp->sval = obstack_finish (_gst_compilation_obstack);
+  lvalp->sval = _gst_obstack_cur_str_buf (_gst_compilation_obstack);
   return (STRING_LITERAL);
 }
 
@@ -620,13 +607,14 @@ scan_ident (int c,
 {
   int ic, identType;
 
-  obstack_1grow (_gst_compilation_obstack, c);
+  _gst_reset_buffer ();
+  _gst_add_str_buf_char (c);
 
   identType = IDENTIFIER;
 
   while (((ic = _gst_next_char ()) != EOF)
 	 && (CHAR_TAB (ic)->char_class & ID_CHAR))
-    obstack_1grow (_gst_compilation_obstack, ic);
+    _gst_add_str_buf_char (ic);
 
   /* Read a dot as '::' if followed by a letter.  */
   if (ic == '.')
@@ -650,7 +638,7 @@ scan_ident (int c,
 	_gst_unread_char (':');
       else
 	{
-          obstack_1grow (_gst_compilation_obstack, ':');
+          _gst_add_str_buf_char (':');
           identType = KEYWORD;
 	}
     }
@@ -658,40 +646,28 @@ scan_ident (int c,
   else
     _gst_unread_char (ic);
 
-  obstack_1grow (_gst_compilation_obstack, '\0');
-  lvalp->sval = obstack_finish (_gst_compilation_obstack);
+  lvalp->sval = _gst_obstack_cur_str_buf (_gst_compilation_obstack);
   return (identType);
 }
 
 
 long double
-mul_powl (long double mant, 
-	  long double base,
-	  int n)
+ipowl (long double base, 
+       int n)
 {
-  long double result = 1.0;
-  int n_abs = n < 0 ? -n : n;
-  int exp;
   int k = 1;
-
-  /* Restrict base to the range from 1.0 to 2.0.  */
-  base = frexpl (base, &exp);
-  base *= 2;
-  exp--;
-
-  while (n_abs)
+  long double result = 1.0;
+  while (n)
     {
-      if (n_abs & k)
+      if (n & k)
         {
           result *= base;
-          n_abs ^= k;
+          n ^= k;
         }
       base *= base;
       k <<= 1;
     }
-
-  mant = (n < 0) ? mant / result : mant * result;
-  return ldexpl (mant, exp * n);
+  return result;
 }
 
 int
@@ -715,9 +691,6 @@ scan_number (int c,
       ic = _gst_next_char ();
       if (ic == 'r')
 	{
-          char *p = obstack_finish (_gst_compilation_obstack);
-          obstack_free (_gst_compilation_obstack, p);
-
 	  base = (int) num;
 	  if (base > 36 || largeInteger)
 	    {
@@ -863,8 +836,10 @@ scan_number (int c,
   else
     _gst_unread_char (ic);
 
-  if (exponent != 0)
-    num = mul_powl (num, (long double) base, exponent);
+  if (exponent > 0)
+    num *= ipowl ((long double) base, exponent);
+  else
+    num /= ipowl ((long double) base, -exponent);
 
   if (isNegative)
     num = -num;
@@ -875,8 +850,6 @@ scan_number (int c,
 
   if (float_type)
     {
-      char *p = obstack_finish (_gst_compilation_obstack);
-      obstack_free (_gst_compilation_obstack, p);
       lvalp->fval = num;
       return (float_type);
     }
@@ -885,7 +858,6 @@ scan_number (int c,
       lvalp->boval = scan_large_integer (isNegative, base);
       return (LARGE_INTEGER_LITERAL);
     }
-  else
     {
       lvalp->ival = (intptr_t) num;
       return (INTEGER_LITERAL);
@@ -901,6 +873,9 @@ scan_digits (int c,
   long double result;
   mst_Boolean oneDigit = false;
 
+  if (largeInteger)
+    _gst_reset_buffer ();
+
   while (c == '_')
     c = _gst_next_char ();
 
@@ -911,7 +886,7 @@ scan_digits (int c,
       result += digit_to_int (c, base);
       if (largeInteger)
 	{
-	  obstack_1grow (_gst_compilation_obstack, digit_to_int (c, base));
+	  _gst_add_str_buf_char (digit_to_int (c, base));
 	  if (result > -MIN_ST_INT
 	      || (!negative && result > MAX_ST_INT))
 	    *largeInteger = true;
@@ -955,7 +930,7 @@ scan_fraction (int c,
 
       if (largeInteger)
         {
-          obstack_1grow (_gst_compilation_obstack, digit_to_int (c, base));
+          _gst_add_str_buf_char (digit_to_int (c, base));
           if (num > MAX_ST_INT)
             *largeInteger = true;
         }
@@ -1031,14 +1006,13 @@ scan_large_integer (mst_Boolean negative,
   gst_uchar *digits, *result;
   byte_object bo;
 
-  /* Copy the contents of the currently grown obstack on the stack.  */
-  size = obstack_object_size (_gst_compilation_obstack);
+  size = _gst_buffer_size ();
   digits = (gst_uchar *) alloca (size);
-  memcpy (digits, obstack_base (_gst_compilation_obstack), size);
+  _gst_copy_buffer (digits);
 
-  /* And reuse the area on the obstack for a struct byte_object.  */
-  obstack_blank (_gst_compilation_obstack, sizeof (struct byte_object));
-  bo = (byte_object) obstack_finish (_gst_compilation_obstack);
+  bo =
+    (byte_object) obstack_alloc (_gst_compilation_obstack,
+				 sizeof (struct byte_object) + size);
 
   bo->class =
     negative ? _gst_large_negative_integer_class :
