@@ -7,7 +7,7 @@
 
 /***********************************************************************
  *
- * Copyright 2003 Free Software Foundation, Inc.
+ * Copyright 2003, 2006 Free Software Foundation, Inc.
  * Written by Paolo Bonzini.
  *
  * This file is part of GNU Smalltalk.
@@ -17,6 +17,28 @@
  * Software Foundation; either version 2, or (at your option) any later 
  * version.
  * 
+ * Linking GNU Smalltalk statically or dynamically with other modules is
+ * making a combined work based on GNU Smalltalk.  Thus, the terms and
+ * conditions of the GNU General Public License cover the whole
+ * combination.
+ *
+ * In addition, as a special exception, the Free Software Foundation
+ * give you permission to combine GNU Smalltalk with free software
+ * programs or libraries that are released under the GNU LGPL and with
+ * independent programs running under the GNU Smalltalk virtual machine.
+ *
+ * You may copy and distribute such a system following the terms of the
+ * GNU GPL for GNU Smalltalk and the licenses of the other code
+ * concerned, provided that you include the source code of that other
+ * code when and as the GNU GPL requires distribution of source code.
+ *
+ * Note that people who make modified versions of GNU Smalltalk are not
+ * obligated to grant this special exception for their modified
+ * versions; it is their choice whether to do so.  The GNU General
+ * Public License gives permission to release a modified version without
+ * this exception; this exception also makes it possible to release a
+ * modified version which carries forward this exception.
+ *
  * GNU Smalltalk is distributed in the hope that it will be useful, but WITHOUT
  * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or 
  * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
@@ -66,7 +88,7 @@ check_against_permission (OOP permissionOOP,
 			  OOP actionOOP)
 {
   gst_permission perm = (gst_permission) OOP_TO_OBJ (permissionOOP);
-  mst_Object actionArray = OOP_TO_OBJ (perm->actions);
+  gst_object actionArray = OOP_TO_OBJ (perm->actions);
   if (perm->name != nameOOP)
     return (false);
 
@@ -241,7 +263,7 @@ mst_Boolean _gst_check_permission (OOP contextOOP,
       for (i = 0; i < num_attributes; i++)
 	{
 	  gst_message attr = (gst_message) OOP_TO_OBJ (info->attributes[i]);
-	  mst_Object attr_args;
+	  gst_object attr_args;
 	  OOP permissionOOP;
 
 	  if (attr->selector != _gst_permission_symbol)
