@@ -201,17 +201,20 @@ call_argtype_function (struct printf_info *const pinfo, int **argtypes, spec_ent
       pinfo->type = spec->type;
 
       if (pinfo->argc > argindex)
-        n = spec->arg(pinfo, pinfo->argc - argindex, *argtypes + argindex);
+        n = spec->arg(pinfo, (size_t) (pinfo->argc - argindex),
+		      *argtypes + argindex);
       else
-        n = spec->arg(pinfo, 0, NULL);
+        n = spec->arg(pinfo, (size_t) 0, NULL);
 
       if (n < 0)
 	return n;
-      else if (argindex + n > pinfo->argc)
+      if (argindex + n > pinfo->argc)
         {
-	  *argtypes = snv_renew (int, *argtypes, argindex + n);
+          int new_ct = argindex + n;
+	  *argtypes = snv_renew (int, *argtypes, new_ct);
+          memset(*argtypes + pinfo->argc, PA_UNKNOWN,
+                 (new_ct - pinfo->argc) * sizeof(**argtypes));
 	  pinfo->argc = argindex + n;
-
 	  /* Call again... */
 	  pinfo->argindex = save_argindex;
 	  pinfo->format = save_format;
@@ -219,7 +222,7 @@ call_argtype_function (struct printf_info *const pinfo, int **argtypes, spec_ent
           pinfo->spec = *pinfo->format;
           pinfo->extra = spec->user;
           pinfo->type = spec->type;
-          n = spec->arg(pinfo, n, *argtypes + argindex);
+          n = spec->arg(pinfo, (size_t)n, *argtypes + argindex);
 	}
     }
 
@@ -395,7 +398,8 @@ parse_printf_format (const char *format, int n, int *argtypes)
 		      info.spec = *info.format;
 		      info.extra = spec->user;
 		      info.type = spec->type;
-		      status = (*spec->arg) (&info, n - argindex, argtypes + argindex);
+		      status = (*spec->arg) (&info, (size_t) (n - argindex),
+					     argtypes + argindex);
 		    }
 		  else
 		    {
@@ -497,7 +501,7 @@ do_printfv (STREAM *stream, const char *format, union printf_arg const args[])
 		  info.extra = spec->user;
 		  info.type = spec->type;
 
-		  status = spec->arg ? (*spec->arg) (&info, 0, NULL) : 1;
+		  status = spec->arg ? (*spec->arg) (&info, (size_t)0, NULL) : 1;
 
 		  if (status < 0)
 		    goto error;
@@ -641,18 +645,36 @@ stream_printfv (STREAM *stream, const char *format, snv_constpointer const *ap)
       /* We scanned the format string to find the type of the arguments,
          so we can now cast it and store it correctly.  */
       for (index = 0; index < info.argc; index++)
-       switch (argtypes[index] & ~PA_FLAG_UNSIGNED)
+        {
+          int tp = argtypes[index];
+          if ((tp & PA_TYPE_MASK) == PA_TYPE_MASK)
+            {
+              if (index + 1 == info.argc)
+                {
+                  info.argc--;
+                }
+              else
+                {
+                  info.error = EINVAL;
+                  index = info.argc;
+                }
+              break;
+            }
+
+          switch (tp & ~PA_FLAG_UNSIGNED)
          {
           case PA_CHAR:
            args[index].pa_char = (char) *(const long int *)(ap + index);
             break;
 
           case PA_WCHAR:
-           args[index].pa_wchar = (snv_wchar_t) *(const long int *)(ap + index);
+              args[index].pa_wchar =
+                (snv_wchar_t) *(const long int *)(ap + index);
             break;
 
           case PA_INT|PA_FLAG_SHORT:
-           args[index].pa_short_int = (short int) *(const long int *)(ap + index);
+              args[index].pa_short_int =
+                (short int) *(const long int *)(ap + index);
             break;
 
           case PA_INT:
@@ -703,6 +725,7 @@ stream_printfv (STREAM *stream, const char *format, snv_constpointer const *ap)
             break;
          }
     }
+    }
 
   if (printf_last_error)
     snv_delete (printf_last_error);
@@ -944,7 +967,8 @@ snv_fdputc (int ch, STREAM *stream)
 {
   static char buf[1] = { 0 };
   buf[0] = (char) ch;
-  return write (SNV_POINTER_TO_INT (stream_details (stream)), buf, 1) ? ch : -1;
+  return write ((int) SNV_POINTER_TO_INT (stream_details (stream)), buf, 1)
+	 ? ch : -1;
 }
 
 /**
